\\ Copyright 2015, 2017, 2018, 2020 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file.  If not, see <http://www.gnu.org/licenses/>.


read("c-curve-defines.gp");
default(realprecision,10);

check_count = 0;
check_exact(got,want, name="") =
{
  check_count++;
  if(got!= want,
     print("\nwrong "name);
     print("got  "got);
     print("want "want);
     quit(1),
     print("ok "check_count));
}

vector_addmul_at(v, a, mul, pos) = for(i=1,#a, v[pos+i-1] += mul*a[i]); v;
r18 = poly18_r18;


\\-----------------------------------------------------------------------------

V1s3_recurrence_factors = {[
  5, -12, 21, -28, 25, -8, -19, 53, -90, 
  112, -118, 108, -88, 48, -16, -8, 32, -32
]};

V1s3_upper(k) = r18^k;
V1s3_lower(k) = r18^k;
V1s3_upper(k) = 1.319373 * r18^k;
V1s3_lower(k) = 1.319372 * r18^k;
{
  my(k = 1);
  upper = 0;
  lower = 0;
  for(i=1,#V1s3_recurrence_factors,
     my(u = V1s3_upper(k-i),
        l = V1s3_lower(k-i),
        f = V1s3_recurrence_factors[i]);
     if(f > 0,
        lower += f*l;
        upper += f*u;
        ,
        lower += f*u;
        upper += f*l));
  upper /= r18^k;
  lower /= r18^k;
  print("upper "upper*1.0);
  print("lower "lower*1.0);
}
quit();

\\-----------------------------------------------------------------------------
\\ r18

\\ 5*a*r^k - 12*b*r^(k-1)
\\ (5*a*r - 12*b) * r^(k-1)
\\ a > 1.8*b
\\ > (5*1.8*b*r - 12*b) * r^(k-1)

\\ 5*a - 12*b    a > 1.8*b
\\ > 5*1.8*b - 12*b
\\ = -3*b

r=1.8;
{
  5 - 12/r + 21/r^2 -28/r^3 + 25/r^4 -8/r^5 -19/r^6 + 53/r^7 -90/r^8
  + 112/r^9 -118/r^10 + 108/r^11 -88/r^12 + 48/r^13 -16/r^14 -8/r^15
  + 32/r^16 -32/r^17
}
quit();


\\-----------------------------------------------------------------------------
\\ (x-1)^2 = x^2 - 2x + 1
\\ f(k+2) - 2*f(k+1) + f(k) = 0
\\ -f(k+2) + 2*f(k+1) = f(k)
\\ if f(k+2) > f(k+1) then 
\\ f(k) = -f(k+2) + 2*f(k+1) < -f(k+1) + 2*f(k+1) = f(k+1)
\\ reversal is descending

f(k) = if(k<=1,k, 2*f(k-1) - f(k-2));
vector(10,k,k--; f(k))

u = concat(-1,V1s3_recurrence_factors);
v = concat(0,u);

vector_addmul_at(v,u,1, 1)
quit();


\\-----------------------------------------------------------------------------

default(realprecision,20);
roots = abs(polroots_descending(poly18(x)))
prod(i=1,#roots, (x-roots[i]))
quit();


\\-----------------------------------------------------------------------------

p18one(k) = {
  if(k<=16, 0,
     k==17, -1,
     sum(i=1,#V1s3_recurrence_factors,
         V1s3_recurrence_factors[i] * p18one(k-i)));
}
vector(30,k,k--; p18one(k))
quit();


\\-----------------------------------------------------------------------------

roots = polroots_descending(poly18(x));
for(i=1,#roots, if(abs(imag(roots[i]))<1e-10,roots[i]=real(roots[i])));
for(i=1,#roots, printf("%30s  abs %s\n", roots[i],abs(roots[i])));

m = matrix(18,18, i,j, roots[j]^(i-1));
v = vectorv(18,i, V1s3(i));
f = matsolve(m,v);

for(i=1,#f, if(abs(imag(f[i]))<1e-10,f[i]=real(f[i])));
print();
{
  for(i=1,#roots,
     printf("(%-35s) * %s ^k\n", f[i], roots[i]));
}
quit();

\\-----------------------------------------------------------------------------

{
  for(k=0,300,
     print(V1s3(k) / r18^k));
}
quit();

\\-----------------------------------------------------------------------------

{
  lo = 18/10;
  hi = 2;
  pos = 0;
  neg = 0;
  for(i=1,#V1s3_recurrence_factors,
        my(a = if(V1s3_recurrence_factors[i] > 0,lo,hi),
           e = -i,
           term = V1s3_recurrence_factors[i]*a^e);
     print(V1s3_recurrence_factors[i]," * "a"^",e," = "term);
     if(V1s3_recurrence_factors[i] > 0, pos+=term, neg+=term);
  );
  print("pos "pos"  "pos*1.0);
  print("neg "neg"  "neg*1.0);
}
quit();

{
  sum(i=1,#V1s3_recurrence_factors,
        if(V1s3_recurrence_factors[i] > 0,
           V1s3_recurrence_factors[i]*(6)^-i))*1.0
}
quit();

{
  \\ V1s3(k+1) >= 3/2 * V1s3(k)

  for(k=1,200, if(V1s3(k) > 2^k, print(k)));
}

\\ 2^k > V1s3(k) > 1
\\ 5*2^k > 5*V1s3(k) > 5
\\ -12*2^k < -12*V1s3(k) < -12
\\ total V1s3 > 5 + -12*2^(k-2)
{
  print();
  print("minimum");
  pos = 1;
  neg = 4;
  t = sum(i=1,#V1s3_recurrence_factors,
            my(a = if(V1s3_recurrence_factors[i]>=0,pos,neg),
               e = -i+1,
               term = V1s3_recurrence_factors[i] * a^e);
            print(V1s3_recurrence_factors[i]," * "a"^",e," = "term);
            term);
  print(t);
  print(t*1.0);
}
{
\\ total V1s3 < 5*2^k -12*0
  print();
  print("maximum");
  pos = 4;
  neg = 0;
  t = sum(i=1,#V1s3_recurrence_factors,
            if(V1s3_recurrence_factors[i]<0, 0,
               my(a = if(V1s3_recurrence_factors[i]>=0,pos,neg),
                  e = -i+1,
                  term = V1s3_recurrence_factors[i] * a^e);
               print(V1s3_recurrence_factors[i]," * "a"^",e," = "term);
               term));
  print(t);
  print(t*1.0);
}
quit();

\\-----------------------------------------------------------------------------
\\ ratios

{
  \\ V1s3(k+1) >= 3/2 * V1s3(k)

  for(k=1,20,
     print(V1s3(k+1) / V1s3(k) * 1.0));
}

\\-----------------------------------------------------------------------------

r = concat(-1, V1s3_recurrence_factors);
v = [1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0];
v = vector_addmul_at(v,r,1, 1);
print(vecsum(v)," ",v);
v = vector_addmul_at(v,r,5, 2);
print(vecsum(v)," ",v);
v = vector_addmul_at(v,r,13, 3);
print(vecsum(v)," ",v);
v = vector_addmul_at(v,r,26, 4);
print(vecsum(v)," ",v);
quit();


