\\ Copyright 2014, 2015 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file.  If not, see <http://www.gnu.org/licenses/>.

allocatemem(300 * 10^6);
default(strictargs,1);
read("pol-pfrac.gp");
want_mirror = 0;

\\-----------------------------------------------------------------------------

join(v,sep) = concat(vector(#v,i,if(i<#v,Str(v[i],sep),v[i])));
conf_to_string(conf) = join(conf,",");

conf_print(conf) =
{
  print("     .",if(conf[9],"--->","    "),".");
  print("    ",
        if(conf[7],"|"," "),
        if(conf[8],"^"," "),
        "    ",
        if(conf[10],"|"," "),
        if(conf[11],"^"," "));
  for(i=1,1,
     print("    ",
           if(conf[7],"|"," "),
           if(conf[8],"|"," "),
           "    ",
           if(conf[10],"|"," "),
           if(conf[11],"|"," ")));
  print("    ",
        if(conf[7],"v"," "),
        if(conf[8],"|"," "),
        "    ",
        if(conf[10],"v"," "),
        if(conf[11],"|"," "));

  print(if(conf[6],"<---","    ")," .",
        if(conf[1],"<---","    "),". ",
        if(conf[12],"<---","    "),
        "  ",conf_to_string(conf),
        ", sides=",conf_num_sides(conf),
        if(conf_is_double_seg(conf),", double",""));
  print(if(conf[5],"--->","    "),"  ",
        if(conf[2],"--->","    "),"  ",
        if(conf[13],"--->","    "));

  print("    ",
        if(conf[4],"|"," "),
        if(conf[3],"^"," "),
        "    ",
        if(conf[15],"|"," "),
        if(conf[14],"^"," "));
  for(i=1,1,
     print("    ",
           if(conf[4],"|"," "),
           if(conf[3],"|"," "),
           "    ",
           if(conf[15],"|"," "),
           if(conf[14],"|"," ")));
  print("    ",
        if(conf[4],"v"," "),
        if(conf[3],"|","."),
        "    ",
        if(conf[15],"v"," "),
        if(conf[14],"|","."));
  print("-----------------------------------------------");
}

\\ conf_num_sides(conf) = \
\\   if(conf[1], 1 + conf[8] + conf[9] + conf[10],  0);
conf_num_sides(conf) = \
  conf[1] + conf[8] + conf[9] + conf[10];
conf_is_single_seg(conf) = conf[1] && ! conf[2];
conf_is_double_seg(conf) = conf[1] && conf[2];
conf_is_empty(conf) = vecsum(conf)==0;

conf_all_ones = [1,1,1,1,1, 1,1,1,1,1, 1,1,1,1,1];
conf_print(conf_all_ones)

\\-----------------------------------------------------------------------------

\\ Numbering of Duvall and Keesling
\\
\\           *---------*                      *---------*
\\          /|\  9|9  /|\                    /|\   |   /|\
\\         / | \  |  / | \                 7/ | \8 |10/ | \11
\\        /  |  \ | /  |  \                /  | 9\ | /9 |  \
\\       /  7|8  \ / 10|11 \              /   |   \ /   |   \
\\      *--- | ---*--  |  --*            *--- | ---*--  |  --*
\\     / \  7|8  / \ 10|11 / \          / \   |   / \   |   / \
\\    /   \  |  /   \  |  /   \       6/   \6 | 1/   \1 |12/   \12
\\   /  |  \ | /  |  \ | /  |  \      /  | 7\ | /8 |10\ | /11|  \
\\  /  6|6  \|/  1|1  \|/ 12|12 \    /   |   \|/   |   \|/   |   \
\\ *---------*---------*---------*  *---------*---------*---------*
\\  \  5|5  /|\  2|2  /|\ 13|13 /    \   |   /|\   |   /|\   |   /
\\   \  |  / | \  |  / | \  |  /      \  | 4/ | \3 |15/ | \14|  /
\\    \   /  |  \   /  |  \   /       5\   /5 | 2\   /2 |13\   /13
\\     \ /  4|3  \ / 15|14 \ /          \ /   |   \ /   |   \ /
\\      *--  |  --*--  |  --*            *--  |  --*--  |  --*
\\       \  4|3  / \ 15|14 /              \   |   / \   |   /
\\        \  |  /   \  |  /                \  |  /   \  |  /
\\         \ | /     \ | /                 4\ | /3  15\ | /14
\\          \|/       \|/                    \|/       \|/
\\           *         *                      *         *
\\
\\ 6 -> left 2bit 2,9
\\

\\ 2: 2,4,6,8 -> 11,13,15,1
\\ 4: 9,7,-1,-1 -> 3,-1,-1,-1
\\ 6: 10,-1,-1,-1 -> 10,-1,-1,-1
\\ 8: 11,13,15,1 -> -1,-1,9,7
\\
\\ 1:1 next 8,6,4,2  look for the 8:0 is 1,15,13,11

\\ expanded    1  2  3  4  5   6  7 8    9 19 11  12,13,14,15
left_perm  = [ 8, 1, 9, 8,10,  9, 1,10, 15, 3, 2,  5, 4, 7, 6; \
               0, 1, 0, 1, 0,  1, 0, 1,  0, 1, 0,  1, 0, 1, 0; \
               4, 0, 6, 0, 8,  0, 2, 0, -1, 0,-1,  0,-1, 0,-1 ];

right_perm = [10, 1,12,11,14, 13, 2,15,  3, 8, 1,  9, 8,10, 9; \
               1, 0, 1, 0, 1,  0, 1, 0,  1, 0, 1,  0, 1, 0, 1; \
               0,11, 0,-1, 0, -1, 0,-1,  0,13, 0, 15, 0, 1,11 ];

perms_vector = [left_perm,right_perm];
perms_name_vector = ["left","right"];
conf_permute(conf,perm) =
{
  my(new_conf = conf);
  for(i=1,#new_conf,
     my(orig_pos = perm[1,i]);
     my(orig_to = conf[orig_pos]);
     if(orig_to==0, new_conf[i]=0; next());

     my(bit=perm[2,i]);
     if(bit==0,new_conf[i] = perm[3,i]; next());

     if(orig_to==-1, new_conf[i]=-1; next());

     new_conf[i] = -1;
     for(j=1,#conf,
        if(perm[1,j]==orig_to && perm[2,j]==0,
           new_conf[i] = j)));
  new_conf;
}

{
  my(allowed = [2,4,6,8;     \\ 1
                1,15,13,11;  \\ 2
                2,4,6,8;     \\ 3
                3,0,0,0;     \\ 4
                2,4,6,8;     \\ 5
                5,0,0,0;     \\ 6
                2,4,6,8;     \\ 7
                7,9,0,0;     \\ 8
                10,0,0,0;    \\ 9
                1,11,13,15;  \\ 10
                10,0,0,0;    \\ 11
                1,11,13,15;  \\ 12
                12,0,0,0;    \\ 13
                1,11,13,15;  \\ 14
                14,0,0,0 ]); \\ 15
  conf_validate(conf) =
  for(i=1,#conf,
     my(to = conf[i]);
     if(to<=0,next());
     for(j=1,4,
        if(allowed[i,j] == to, next(2)));
     print("conf ",conf_to_string(conf)," invalid pos="i" to "to" should be "allowed[i,]);
     conf_print(conf);
     quit(1));
}

\\---------

\\ c=vector(15);
\\ c[6+1] = 1;
\\ c[9+1] = 1;
\\ c = conf_permute(c,left_perm);
\\ c
\\ conf_print(c)
\\ quit

conf_initial_vector = vector(15,c, vector(15,i, -(i==c)));
\\ conf_initial[6+1] = 1;
\\ conf_initial
\\ print(conf_to_string(conf_initial));
\\ for(i=1,#conf_initial_vector, conf_print(conf_initial_vector[i]));

\\ conf_permute(conf_initial,left_perm)
\\ conf_permute(conf_initial,right_perm)
\\ conf_print(conf_permute(conf_initial,left_perm));
\\ conf_print(conf_permute(conf_initial,right_perm));

conf_print(conf_initial_vector[1])
{
  my(conf = conf_initial_vector[1]);
  my(l = conf_permute(conf,left_perm));
  my(r = conf_permute(conf,right_perm));
  conf_print(l);
 \\ conf_print(r);
  my(ll = conf_permute(l,left_perm));
  my(lr = conf_permute(l,right_perm));
  conf_print(ll);
  conf_print(lr);
}


\\-------------------------------------

conf_to_index_locked = 0;
conf_to_index_key   = List([]);
conf_to_index_value = List([]);
index_to_conf = List([]);
conf_to_index(conf) =
{
  my(pos=setsearch(conf_to_index_key,conf));
  if(pos>0, return(conf_to_index_value[pos]));

  if(conf_to_index_locked,
     error("Unknown conf ",conf_to_string(conf)));

  conf_validate(conf);

  listput(index_to_conf,conf);
  my(index = #index_to_conf);

  pos = setsearch(conf_to_index_key,conf,1);  \\ where to insert
  listinsert(conf_to_index_key,  conf,  pos);
  listinsert(conf_to_index_value,index, pos);

  \\ when treating mirrors as equivalent insert also mirror image conf
  if(want_mirror,
     conf = conf_permute(conf,mirror_perm);
     pos = setsearch(conf_to_index_key,conf,1); \\ where to insert
     if(pos,
        listinsert(conf_to_index_key,  conf,  pos);
        listinsert(conf_to_index_value,index, pos)));

  index;
}

{
  for(i=1,#conf_initial_vector, conf_to_index(conf_initial_vector[i]));
  my(i=1,
     verbose=0);
  while(i <= #index_to_conf,
        my(conf=index_to_conf[i]);
        if(verbose, print(conf_to_string(conf)));
        i++;
        for(p=1,#perms_vector,
           my(perm=perms_vector[p],
              new_conf=conf_permute(conf,perm));
           if(verbose, print("  ",perms_name_vector[p],
                             "->  ",conf_to_string(new_conf)));
           if(conf_is_empty(new_conf),
              next());
           conf_to_index(new_conf)));
}
conf_to_index_locked = 1;
index_to_conf = Vec(index_to_conf);

for(i=1,#index_to_conf, conf_validate(index_to_conf[i]));
print("conf_validate ok");

size = #index_to_conf;
print("configurations size=",size);

m = matrix(size,size);
m_lr=matrix(size,2);
{
for(index=1,size,
   my(conf = index_to_conf[index]);
   for(p=1,#perms_vector,
      my(perm      = perms_vector[p],
         new_conf  = conf_permute(conf,perm));
      if(conf_is_empty(new_conf),
         print("skip ",conf_to_string(conf)," "perms_name_vector[p]" -> empty");
         next());
      my(new_index = conf_to_index(new_conf));
      m[new_index,index]++;
      m_lr[index,p] = new_index;
   ));
}

mat_sum(m,f) = my(s=matsize(m)); sum(i=1,s[1], sum(j=1,s[2], f(m[i,j])));
mat_count_nonzeros(m) = mat_sum(m,(x)->x!=0);
mat_count_equal(m,value) = mat_sum(m,(x)->x==value);

print("m determinant ",matdet(m));
print("m total weight ", sum(c=1,size, vecsum(m[,c])));
print("m non-zeros    ",mat_count_nonzeros(m));
print("m twos         ",mat_count_equal(m,2));
initial = vectorv(size,i, i==1);

initial = vectorv(size,i, 0);
for(i=1,#conf_initial_vector, \
    initial[conf_to_index(conf_initial_vector[i])]++);
\\ print("initial "initial)

{
for(i=1,#index_to_conf, \
   my(conf = index_to_conf[i],
      conf_left = conf_permute(conf,left_perm));
   if(conf_left == conf_permute(conf,right_perm),
      print("conf ",conf_to_string(conf),
            " left,right = ",conf_to_string(conf_left))));
}

\\----------------------
\\ print conf descents

\\ {
\\ my(v=initial);
\\ for(k=0,5,
\\    print("confs k=" k);
\\    for(i=1,#v, if(v[i], conf_print(index_to_conf[i])));
\\    for(i=1,#v, if(v[i],
\\                   my(conf=index_to_conf[i]);
\\                   print(conf_to_string(conf),
\\                         " -> ",
\\                         conf_to_string(conf_permute(conf,left_perm)),
\\                         " "
\\                         conf_to_string(conf_permute(conf,right_perm)))));
\\    v = m*v);
\\ }

\\-----------------------------------------------------------------------------

{
  my(count=0);
  for(i=1,size,
     for(j=1,size,
        if(m[i,j]>1,
           print(i","j" m="m[i,j]);
           count++)));
  print("total " count " doubles");
}
{
  my(maxdepth = 18,
     reached  = matrix(maxdepth,size,depth,i, depth==1));
  for(depth=2,maxdepth,
     for(i=1,size,
        if(reached[depth-1,i],
           for(j=1,2,
              my(r=m_lr[i,j]); if(r, reached[depth,r] = 1)))));
  my(prev_unreached=0);
  for(depth=1,maxdepth,
     my(count_reached   = vecsum(reached[depth,]),
        count_unreached = size - count_reached,
        new_unreached   = count_unreached - prev_unreached);
     print("depth ",depth,
           " reached ",count_reached,
           " unreached ",count_unreached,
           " new ",new_unreached);
     if(depth>1 && new_unreached==0,break());
     prev_unreached = count_unreached);
}

\\-----------------------------------------------------------------------------

conf_index_single_destination(index) =
{
  if(m_lr[index,1]==0, m_lr[index,2],
     m_lr[index,2]==0, m_lr[index,1],
     0);
}
{
  print("single destinations");
  for(i=1,#index_to_conf,
     my(new_index = conf_index_single_destination(i));
     if(new_index,
        my(conf = index_to_conf[i],
           new_conf = index_to_conf[new_index],
           new_single = conf_index_single_destination(new_index));
        print(i "   ",conf_to_string(conf),
              " -> ",
              new_index,"  ",conf_to_string(new_conf)
              \\ , " [" new_single "]"
             )));
}
\\ 11 -> 4 -> 13 -> 6 -> 15=not
\\ 7 -> 14 -> 5 -> 12 -> 3=not
\\ 10 -> 22=not
\\ 23 -> 30 -> 40 -> 20=not
\\ 24 -> 31 -> 41 -> 21=not
\\ 32 -> 42=not
\\ 45 -> 58=not
\\ 84 -> 110=not
\\ 89 -> 116=not

\\ {
\\   my(unseen=vector(size,i,1));
\\   for(i=1,size,
\\      my(p=m_lr[i,1]);
\\      if(!unseen[p], m_lr[i,1]=m_lr[i,2]; m_lr[i,2]=p); \\ swap
\\      unseen[m_lr[i,1]]=0);
\\   print("unseen ",vecsum(unseen));
\\ }
\\ {
\\   my(unseen=vector(size,i,1));
\\   for(i=1,size, unseen[m_lr[i,1]]=0; unseen[m_lr[i,2]]=0);
\\   print("unseen ",vecsum(unseen));
\\ }


\\-----------------------------------------------------------------------------
s_matrix(n) = my(m=matrix(size,size)); m[1,n]=1; m;

\\ C*s shifts row 1 to row n, all others 0
column_1_to_n(n) = my(m=matrix(size,size)); m[n,1]=1; m;

\\  1 0 0 0    0 1 0 0    0 0 1 0    0 0 0 1
\\  0 0 0 0    0 0 0 0    0 0 0 0    0 0 0 0
\\  0 0 0 0    0 0 0 0    0 0 0 0    0 0 0 0
\\  0 0 0 0    0 0 0 0    0 0 0 0    0 0 0 0
\\
\\  0 0 0 0    0 0 0 0    0 0 0 0    0 0 0 0
\\  1 0 0 0    0 1 0 0    0 0 1 0    0 0 0 1
\\  0 0 0 0    0 0 0 0    0 0 0 0    0 0 0 0
\\  0 0 0 0    0 0 0 0    0 0 0 0    0 0 0 0
\\
\\  0 0 0 0
\\  1 0 0 0
\\  0 0 0 0
\\  0 0 0 0

\\ row 1 from each power
\\ T = S0*m^0 + S1*m^1 + S2*m^2 + S3*m^3
\\ C*T * m * (C*T)^-1

\\-----------------------------------------------------------------------------

\\ print("matsolve\n");
\\ mx = matid(size) - m*x;
\\ gm = matsolve(mx,initial);
\\ gm
\\ index_to_conf
