\\ Copyright 2015, 2017 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file.  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.

read("../tools/memoize.gp");

atan2(y,x) = my(angle=arg(x+I*y)); if(angle<0,angle+2*Pi,angle);
deg(x) = x/180*Pi;

nearly_equal(x,y) = abs(x-y) <= max(abs(x),max(abs(y),1))*1e-10;

check_count = 0;
differ(x,y) =
{
  \\ if(type(x)!=type(y),
  \\    return(Strprintf("different types %s %s",type(x),type(y))));
  if(type(x)=="t_COMPLEX" || type(y)=="t_COMPLEX",
     my(d=differ(real(x),real(y)));
     if(d != 0, return(Strprintf("real %s",d)));
     d=differ(imag(x),imag(y));
     if(d != 0, return(Strprintf("imag %s",d)));
     return(0));

  if(type(x)=="t_VEC" || type(x)=="t_COL",
     if(length(x)!=length(y),
        return(Strprintf("different lengths %s %s",length(x),length(y))));
     for(i=1,#x,
        my(d=differ(x[i],y[i]));
        if(d != 0, return(Strprintf("at index %d %s",i,d))));
     return(0));

  if(! nearly_equal(x,y),
     return(Strprintf("different values\n%s\n%s",x,y)));
  return(0);
}

check_equal(got,want, name="") =
{
  if(name!="", name=Str(" -- ",name));
  my(d = differ(got,want));
  if(d != 0,
     print("wrong\n",d); quit(1),
     print("ok",name));
}

check_exact(got,want, name="") =
{
  check_count++;
  if(got!= want,
     print("\nwrong "name);
     print("got  "got);
     print("want "want);
     quit(1),
     print("ok "check_count));
}

\\-------------

Iz(k) =  1/2*4^k - (1/2*k + 1/4)*2^k - 1/4;

b=1+I;
c_curve_segments(k) = {
  my(v = [1/2], end=1);
  for(i=1,k,
      v = concat(v, apply((z)->end+z*I, v));
      end *= b);
  v;
}

Ix_by_segments(k) =
{
  my(v = c_curve_segments(k));
  my(centroid = vecsum(v)/#v);
  v = apply(z->(z-centroid), v);
  v *= 1/(b/abs(b))^(k-2);
  sum(i=1,#v, real(v[i])^2);
}
Ix_by_segments = memoize(Ix_by_segments);

Iy_by_segments(k) =
{
  my(v = c_curve_segments(k));
  my(centroid = vecsum(v)/#v);
  v = apply(z->(z-centroid), v);
  v *= 1/(b/abs(b))^(k-2);
  sum(i=1,#v, imag(v[i])^2);
}
Iy_by_segments = memoize(Iy_by_segments);

Ixy_by_segments(k) =
{
  my(v = c_curve_segments(k));
  my(centroid = vecsum(v)/#v);
  v = apply(z->(z-centroid), v);
  v *= 1/(b/abs(b))^(k-2);
  sum(i=1,#v, real(v[i])*imag(v[i]));
}
Ixy_by_segments = memoize(Ixy_by_segments);

Iz_by_segments(k) =
{
  my(v = c_curve_segments(k));
  my(centroid = vecsum(v)/#v);
  v = apply(z->(z-centroid), v);
  sum(i=1,#v, norm(v[i]));
}
Iz_by_segments = memoize(Iz_by_segments);

vector(5,k,my(k=k-1); Iz_by_segments(k))
vector(5,k,my(k=k-1); Ix_by_segments(k) + Iy_by_segments(k))
check_equal(vector(10,k,my(k=k-1); Iz_by_segments(k)), \
            vector(10,k,my(k=k-1); Ix_by_segments(k) + Iy_by_segments(k)))

Ika_by_segments(k,a) =
{
  my(v = c_curve_segments(k),
     centroid = vecsum(v)/#v,
     rotate = 1/((b/abs(b))^k * exp(a*I)));
  v = apply(z->(z-centroid)*rotate, v);
  sum(i=1,#v, imag(v[i])^2);
}
Ika_by_segments = memoize(Ika_by_segments);
check_equal(Ika_by_segments(0,0), 0);
check_equal(Ika_by_segments(0,deg(90)), 0);
check_equal(Ika_by_segments(0,deg(10)), 0);

sqrt2 = sqrt(2);
G(k) = 1/2*(1 - 1/2^k);
G_unit_length(k) = G(k)*sqrt2^k;
Iz_t1(k) = G_unit_length(k-1) /sqrt2;
Iz_t2(k) = 1/2 * sqrt2^(k-1) /sqrt2;
Iz_t(k) = Iz_t1(k) +  Iz_t2(k);

\\-------------

check_equal(cos(deg(45)), 1/sqrt(2))
check_equal(vector(256,k,my(a=2*Pi/k); cos(a+deg(45))), \
            vector(256,k,my(a=2*Pi/k); cos(a)*1/sqrt(2) - sin(a)*1/sqrt(2)));
check_equal(vector(256,k,my(a=2*Pi/k); cos(a-deg(45))), \
            vector(256,k,my(a=2*Pi/k); cos(a)*1/sqrt(2) + sin(a)*1/sqrt(2)));

check_equal(vector(256,k,my(a=2*Pi/k); cos(a+deg(45))^2), \
            vector(256,k,my(a=2*Pi/k); 1/2*cos(a)^2 - cos(a)*sin(a) + 1/2*sin(a)^2));
check_equal(vector(256,k,my(a=2*Pi/k); cos(a-deg(45))^2), \
            vector(256,k,my(a=2*Pi/k); 1/2*cos(a)^2 + cos(a)*sin(a) + 1/2*sin(a)^2));

\\-------------

check_equal(vector(256,k,my(a=2*Pi/k); sin(a+deg(45))), \
            vector(256,k,my(a=2*Pi/k); cos(a)*1/sqrt(2) + sin(a)*1/sqrt(2)));
check_equal(vector(256,k,my(a=2*Pi/k); sin(a-deg(45))), \
            vector(256,k,my(a=2*Pi/k); - cos(a)*1/sqrt(2) + sin(a)*1/sqrt(2)));

check_equal(vector(256,k,my(a=2*Pi/k); sin(a+deg(45))^2), \
            vector(256,k,my(a=2*Pi/k); 1/2*cos(a)^2 + cos(a)*sin(a) + 1/2*sin(a)^2));
check_equal(vector(256,k,my(a=2*Pi/k); sin(a-deg(45))^2), \
            vector(256,k,my(a=2*Pi/k); 1/2*cos(a)^2 - cos(a)*sin(a) + 1/2*sin(a)^2));

\\-------------

check_equal(vector(256,k,my(a=2*Pi/k); cos(a+deg(45))*sin(a+deg(45))), \
            vector(256,k,my(a=2*Pi/k);   1/2*cos(a)^2 - 1/2*sin(a)^2));
check_equal(vector(256,k,my(a=2*Pi/k); cos(a-deg(45))*sin(a-deg(45))), \
            vector(256,k,my(a=2*Pi/k); - 1/2*cos(a)^2 + 1/2*sin(a)^2));

\\-------------

Ika = Ika_by_segments;

Ika_by_parallel_axis(k,a) = {
  if(k==0,return(0));
    Ika(k-1, a - deg(45))
  + Ika(k-1, a + deg(45))
  + 2^(k-1) * ( Iz_t(k) * sin(a) )^2         /* m*t^2 */
  + 2^(k-1) * ( Iz_t(k) * sin(a) )^2         /* m*t^2 */
  ;
}

print("Ixy_by_segments        ", vector(6,k,my(k=k-1); Ixy_by_segments(k)))

print("Ix_by_segments         ", vector(4,k,my(k=k-1); Ix_by_segments(k)))
print("Ika_by_parallel_axis X ", vector(4,k,my(k=k-1); Ika_by_segments(k,0)))

print("Iy_by_segments         ", vector(4,k,my(k=k-1); Iy_by_segments(k)))
print("Ika_by_parallel_axis Y ", vector(4,k,my(k=k-1); Ika_by_segments(k,deg(90))))

check_equal(vector(20,a,my(a=2*Pi/a); vector(5,k,my(k=k-1); Ika_by_segments(k,a))), \
            vector(20,a,my(a=2*Pi/a); vector(5,k,my(k=k-1); Ika_by_parallel_axis(k,a))), \
            "Ika_by_parallel_axis")


\\-------------

\\ Ika(k-1, a + 45)
\\ Ix*cos(a)^2 + Ixy*cos*sin + Iy*sin^2
\\ = Ix  * ( 1/2*cos(a)^2 - cos(a)*sin(a) + 1/2*sin(a)^2 )
\\ + Ixy * ( 1/2*cos(a)^2                 - 1/2*sin(a)^2 )
\\ + Iy  * ( 1/2*cos(a)^2 + cos(a)*sin(a) + 1/2*sin(a)^2 )

\\ Ika(k-1, a - 45)
\\ Ix*cos(a)^2 + Ixy*cos*sin + Iy*sin^2
\\ = Ix  * (  1/2*cos(a)^2 + cos(a)*sin(a) + 1/2*sin(a)^2 )
\\ + Ixy * ( -1/2*cos(a)^2                 + 1/2*sin(a)^2 )
\\ + Iy  * (  1/2*cos(a)^2 - cos(a)*sin(a) + 1/2*sin(a)^2 )

{   \\ coeff cos(a)^2
Ix(k)  = if(k==0,return(0));
           1/2* Ix(k-1) + 1/2* Ixy(k-1) + 1/2* Iy(k-1)
         + 1/2* Ix(k-1) - 1/2* Ixy(k-1) + 1/2* Iy(k-1)
         ;
}
{   \\ coeff cos(a)*sin(a)
Ixy(k) = if(k==0,return(0));
         - Ix(k-1) + Iy(k-1)
         + Ix(k-1) - Iy(k-1)
         ;
}
{
Iy(k)  = if(k==0,return(0));
           1/2* Ix(k-1) - 1/2* Ixy(k-1) + 1/2* Iy(k-1)
         + 1/2* Ix(k-1) + 1/2* Ixy(k-1) + 1/2* Iy(k-1)
       + 2^k * Iz_t(k)^2
       ;
}
Ix = memoize(Ix);
Ixy= memoize(Ixy);
Iy = memoize(Iy);

print("Ix_by_segments ", vector(4,k,my(k=k-1); Ix_by_segments(k)))
print("Ix             ", vector(4,k,my(k=k-1); Ix(k)))

print("Iy_by_segments ", vector(4,k,my(k=k-1); Iy_by_segments(k)))
print("Iy             ", vector(4,k,my(k=k-1); Iy(k)))

Ika(k,a) = Ix(k)*cos(a)^2 + Ixy(k)*cos(a)*sin(a) + Iy(k)*sin(a)^2;

check_equal(vector(5,k,my(k=k-1); Ix(k)), \
            vector(5,k,my(k=k-1); Ix_by_segments(k)))
check_equal(vector(5,k,my(k=k-1); Iy(k)), \
            vector(5,k,my(k=k-1); Iy_by_segments(k)))

check_equal(vector(20,a,my(a=2*Pi/a); vector(5,k,my(k=k-1); Ika(k,a))), \
            vector(20,a,my(a=2*Pi/a); vector(5,k,my(k=k-1); Ika_by_segments(k,a))))

\\----------------------

Ix(k)  = if(k==0,return(0));  Ix(k-1) + Iy(k-1);
Ixy(k) = 0;
Iy(k)  = if(k==0,return(0));  Ix(k-1) + Iy(k-1) + 2^k * Iz_t(k)^2;
Ix = memoize(Ix);
Ixy= memoize(Ixy);
Iy = memoize(Iy);

Ika(k,a) = Ix(k)*cos(a)^2 + Ixy(k)*cos(a)*sin(a) + Iy(k)*sin(a)^2;

check_equal(vector(5,k,my(k=k-1); Ix(k)), \
            vector(5,k,my(k=k-1); Ix_by_segments(k)))
check_equal(vector(5,k,my(k=k-1); Iy(k)), \
            vector(5,k,my(k=k-1); Iy_by_segments(k)))
check_equal(vector(5,k,my(k=k-1); Iz(k)), \
            vector(5,k,my(k=k-1); Iz_by_segments(k)))

check_equal(vector(20,a,my(a=2*Pi/a); vector(5,k,my(k=k-1); Ika(k,a))), \
            vector(20,a,my(a=2*Pi/a); vector(5,k,my(k=k-1); Ika_by_segments(k,a))))

\\----------------------

Ix(k)  = if(k==0,return(0));  Iz(k-1);
Ixy(k) = 0;
Iy(k)  = if(k==0,return(0));  Iz(k-1) + 2^k * Iz_t(k)^2;

Ika(k,a) = Ix(k)*cos(a)^2 + Ixy(k)*cos(a)*sin(a) + Iy(k)*sin(a)^2;

check_equal(vector(5,k,my(k=k-1); Ix(k)), \
            vector(5,k,my(k=k-1); Ix_by_segments(k)))
check_equal(vector(5,k,my(k=k-1); Iy(k)), \
            vector(5,k,my(k=k-1); Iy_by_segments(k)))
check_equal(vector(5,k,my(k=k-1); Iz(k)), \
            vector(5,k,my(k=k-1); Iz_by_segments(k)))

check_equal(vector(20,a,my(a=2*Pi/a); vector(5,k,my(k=k-1); Ika(k,a))), \
            vector(20,a,my(a=2*Pi/a); vector(5,k,my(k=k-1); Ika_by_segments(k,a))))

\\----------------------
\\ 2^k*t^2 = 1/4*(4^k - 2*2^k + 1)

Ix(k)  =         1/8*(4^k - (2*k - 1)*2^k - 2);
Iy(k)  = Ix(k) + 1/4*(4^k - 2*2^k + 1);
Iz(k)  = Ix(k+1);

Ika(k,a) = Ix(k)*cos(a)^2 + Iy(k)*sin(a)^2;

check_equal(vector(5,k,my(k=k-1); Ix(k)), \
            vector(5,k,my(k=k-1); Ix_by_segments(k)))
check_equal(vector(5,k,my(k=k-1); Iy(k)), \
            vector(5,k,my(k=k-1); Iy_by_segments(k)))
check_equal(vector(5,k,my(k=k-1); Iz(k)), \
            vector(5,k,my(k=k-1); Iz_by_segments(k)))

check_equal(vector(20,a,my(a=2*Pi/a); vector(5,k,my(k=k-1); Ika(k,a))), \
            vector(20,a,my(a=2*Pi/a); vector(5,k,my(k=k-1); Ika_by_segments(k,a))))

\\----------------------
\\ 2^k*t^2 = 1/4*(4^k - 2*2^k + 1)

Ix(k)  =         1/8*(  4^k - (2*k - 1)*2^k - 2);
Iy(k)  = Ix(k) + 1/8*(2*4^k -        4*2^k  + 2);
Iz(k)  = Ix(k+1);
\\ Iz(k) = 1/2*4^k - (1/2*k + 1/4)*2^k - 1/4

Ika(k,a) = Ix(k)*cos(a)^2 + Iy(k)*sin(a)^2;

check_exact(Ix(0),0);
check_exact(Iy(0),0);
check_exact(Iz(0),0);
check_equal(vector(5,k,my(k=k-1); Ix(k)), \
            vector(5,k,my(k=k-1); Ix_by_segments(k)))
check_equal(vector(5,k,my(k=k-1); Iy(k)), \
            vector(5,k,my(k=k-1); Iy_by_segments(k)))
check_equal(vector(5,k,my(k=k-1); Iz(k)), \
            vector(5,k,my(k=k-1); Iz_by_segments(k)))

check_equal(vector(20,a,my(a=2*Pi/a); vector(5,k,my(k=k-1); Ika(k,a))), \
            vector(20,a,my(a=2*Pi/a); vector(5,k,my(k=k-1); Ika_by_segments(k,a))))

\\----------------------

Ix(k)  = 1/8*(  4^k - (2*k - 1)*2^k - 2);
Iy(k)  = 1/8*(3*4^k - (2*k + 3)*2^k    );
Iz(k)  = 1/8*(4*4^k - (4*k + 2)*2^k - 2);

Ika(k,a) = Ix(k)*cos(a)^2 + Iy(k)*sin(a)^2;

check_exact(Ix(0),0);
check_exact(Iy(0),0);
check_exact(Iz(0),0);
check_equal(vector(5,k,my(k=k-1); Ix(k)), \
            vector(5,k,my(k=k-1); Ix_by_segments(k)))
check_equal(vector(5,k,my(k=k-1); Iy(k)), \
            vector(5,k,my(k=k-1); Iy_by_segments(k)))
check_equal(vector(5,k,my(k=k-1); Iz(k)), \
            vector(5,k,my(k=k-1); Iz_by_segments(k)))

check_equal(vector(20,a,my(a=2*Pi/a); vector(5,k,my(k=k-1); Ika(k,a))), \
            vector(20,a,my(a=2*Pi/a); vector(5,k,my(k=k-1); Ika_by_segments(k,a))))

\\----------------------

Ix(k)  = 1/8*(4^k - (2*k-1)*2^k - 2);
Iy(k)  = Ix(k+1) - Ix(k);
Iz(k)  = Ix(k+1);

Ika(k,a) = Ix(k)*cos(a)^2 + Iy(k)*sin(a)^2;

check_exact(Ix(0),0);
check_exact(Iy(0),0);
check_exact(Iz(0),0);
check_equal(vector(5,k,my(k=k-1); Ix(k)), \
            vector(5,k,my(k=k-1); Ix_by_segments(k)))
check_equal(vector(5,k,my(k=k-1); Iy(k)), \
            vector(5,k,my(k=k-1); Iy_by_segments(k)))
check_equal(vector(5,k,my(k=k-1); Iz(k)), \
            vector(5,k,my(k=k-1); Iz_by_segments(k)))

check_equal(vector(20,a,my(a=2*Pi/a); vector(5,k,my(k=k-1); Ika(k,a))), \
            vector(20,a,my(a=2*Pi/a); vector(5,k,my(k=k-1); Ika_by_segments(k,a))))

\\----------------------

Ix(k)  = sum(i=0,k-1, Iy(i));
Iy(k)  = 1/8*( 3*4^k - (2*k + 3)*2^k );
Iz(k)  = sum(i=0,k, Iy(i));

Ika(k,a) = Ix(k)*cos(a)^2 + Iy(k)*sin(a)^2;

check_exact(Ix(0),0);
check_exact(Iy(0),0);
check_exact(Iz(0),0);
check_equal(vector(5,k,my(k=k-1); Ix(k)), \
            vector(5,k,my(k=k-1); Ix_by_segments(k)))
check_equal(vector(5,k,my(k=k-1); Iy(k)), \
            vector(5,k,my(k=k-1); Iy_by_segments(k)))
check_equal(vector(5,k,my(k=k-1); Iz(k)), \
            vector(5,k,my(k=k-1); Iz_by_segments(k)))

check_equal(vector(20,a,my(a=2*Pi/a); vector(5,k,my(k=k-1); Ika(k,a))), \
            vector(20,a,my(a=2*Pi/a); vector(5,k,my(k=k-1); Ika_by_segments(k,a))))

\\----------------------
vector(10,k, Ix(k-1))*4
vector(10,k, Iy(k-1))*4
vector(10,k, Iz(k-1))*4

\\----------------------
\\ sum

I_2kt_squared(k) = 1/4*(4^k - 2*2^k + 1);

Ix(k)  = sum(j=0,k-1, 2^(k-1-j)*I_2kt_squared(j));
Iy(k)  = if(k==0,return(0)); \
         1/2*sum(j=0,k, 2^(k-j)*I_2kt_squared(j)) + 1/2*I_2kt_squared(k);
Iz(k)  = sum(j=0,k, 2^(k-j)*I_2kt_squared(j));

print("\nnew");
vector(10,k, Ix(k-1))*4
vector(10,k, Iy(k-1))*4
vector(10,k, Iz(k-1))*4

check_equal(vector(5,k,my(k=k-1); Ix(k)), \
            vector(5,k,my(k=k-1); Ix_by_segments(k)))
check_equal(vector(5,k,my(k=k-1); Iy(k)), \
            vector(5,k,my(k=k-1); Iy_by_segments(k)))
check_equal(vector(5,k,my(k=k-1); Iz(k)), \
            vector(5,k,my(k=k-1); Iz_by_segments(k)))

\\----------------------
\\ mutual

Ix(k)  = if(k==0,return(0)); 2*( (1/sqrt2)^2*Ix(k-1) + (1/sqrt2)^2*Iy(k-1) );
Iy(k)  = if(k==0,return(0)); 2*( (1/sqrt2)^2*Ix(k-1) + (1/sqrt2)^2*Iy(k-1) )  + 2^k*Iz_t(k)^2;
Iz(k)  = 1/2*4^k - (1/2*k + 1/4)*2^k - 1/4;

\\ for(i=0,90,printf("%*s",90-i,"");print(decimal_to_binary(Ix(i)*4)))
\\ for(i=0,90,printf("%*s",90-i,"");print(decimal_to_binary(Iy(i)*4)))
\\ for(i=0,90,printf("%*s",90-i,"");print(decimal_to_binary(Iz(i)*4)))

check_exact(Ix(0),0);
check_exact(Iy(0),0);
check_exact(Iz(0),0);
check_equal(vector(5,k,my(k=k-1); Ix(k)), \
            vector(5,k,my(k=k-1); Ix_by_segments(k)))
check_equal(vector(5,k,my(k=k-1); Iy(k)), \
            vector(5,k,my(k=k-1); Iy_by_segments(k)))
check_equal(vector(5,k,my(k=k-1); Iz(k)), \
            vector(5,k,my(k=k-1); Iz_by_segments(k)))

\\----------------------
\\ mutual
\\ Iy(k-1) = Ix(k) - Ix(k-1) 
\\ Iy(k) = Ix(k+1) - Ix(k) 

Ix(k) = if(k<2,return(0)); 2*Ix(k-1)   + 2^(k-1)*Iz_t(k-1)^2;


\\ for(i=0,90,printf("%*s",90-i,"");print(decimal_to_binary(Ix(i)*4)))
\\ for(i=0,90,printf("%*s",90-i,"");print(decimal_to_binary(Iy(i)*4)))
\\ for(i=0,90,printf("%*s",90-i,"");print(decimal_to_binary(Iz(i)*4)))

check_exact(Ix(0),0);
check_exact(Iy(0),0);
check_exact(Iz(0),0);
check_equal(vector(5,k,my(k=k-1); Ix(k)), \
            vector(5,k,my(k=k-1); Ix_by_segments(k)))
check_equal(vector(5,k,my(k=k-1); Iy(k)), \
            vector(5,k,my(k=k-1); Iy_by_segments(k)))
check_equal(vector(5,k,my(k=k-1); Iz(k)), \
            vector(5,k,my(k=k-1); Iz_by_segments(k)))
