\\ Copyright 2015 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file.  If not, see <http://www.gnu.org/licenses/>.


read("../tools/memoize.gp");

S_samples = [ 1,2,4, 8,14,28, 54,108,210,  412, 806, 1588, 3098, 6052, 11822 ];
S(k) = if(k<9,return(S_samples[k+1])); \
       3*S(k-1) \
    -  3*S(k-2) \
    +  3*S(k-3) \
    -  2*S(k-4) \
    -  4*S(k-5) \
    +  8*S(k-6) \
    -  8*S(k-7) \
    + 16*S(k-8) \
    -  8*S(k-9);
S = memoize(S);

DJ_samples = [ 0, 0, 0, 1, 0, 1, 0, 3, 4, 9, 12, 39, 72, 141, 264, 547, 1036 ];
DJ(k) = if(k<9,return(DJ_samples[k+1])); \
       3*DJ(k-1) \
    -  3*DJ(k-2) \
    +  3*DJ(k-3) \
    -  2*DJ(k-4) \
    -  4*DJ(k-5) \
    +  8*DJ(k-6) \
    -  8*DJ(k-7) \
    + 16*DJ(k-8) \
    -  8*DJ(k-9);
DJ = memoize(DJ);

D(k) = if(k==0, 0, 2*D(k-1) + DJ(k-1));
D = memoize(D);

\\ D exceeds S at k=48 onwards
\\ 2*D exceeds S at k=30 onwards
for(k=0,100, print("k="k" "2*D(k)-S(k)));


D_matrix = {[  0,  1, 0,  0, 0,  0, 0, 0, 0, 0;
               0,  0, 1,  0, 0,  0, 0, 0, 0, 0;
               0,  0, 0,  1, 0,  0, 0, 0, 0, 0;
               0,  0, 0,  0, 1,  0, 0, 0, 0, 0;
               0,  0, 0,  0, 0,  1, 0, 0, 0, 0;

               0,  0, 0,  0, 0,  0, 1, 0, 0, 0;
               0,  0, 0,  0, 0,  0, 0, 1, 0, 0;
               0,  0, 0,  0, 0,  0, 0, 0, 1, 0;
               0,  0, 0,  0, 0,  0, 0, 0, 0, 1;
              16,-40,32,-24,16,  0,-8, 9,-9, 5 ]};

poly9(x) = 1 - 3*x + 3*x^2 - 3*x^3 + 2*x^4 + 4*x^5 - 8*x^6 + 8*x^7 - 16*x^8 + 8*x^9;
gS(x) = (1 - x + x^2 - x^3 - 2*x^4 + 6*x^5 - 4*x^6 + 12*x^7 - 8*x^8)/poly9(x);
gDJ(x) = x^3 * (1 - 3*x + 4*x^2 - 6*x^3 + 8*x^4 - 4*x^5) / poly9(x);
gD(x) = x * gDJ(x) / (1-2*x);
gDS(x) = gD(x) - gS(x);
gDS(x)

DS_by_difference(k) = D(k) - S(k);
DS_samples = vectorv(10,k,my(k=k-1); DS_by_difference(k))
DS_by_recurrence(k) = {
  if(k<10,return(DS_samples[k+1]));
    5*DS(k-1)
  - 9*DS(k-2)
  + 9*DS(k-3)
  - 8*DS(k-4)
  + 0*DS(k-5)
  + 16*DS(k-6)
  - 24*DS(k-7)
  + 32*DS(k-8)
  - 40*DS(k-9)
  + 16*DS(k-10);
}
DS_by_recurrence = memoize(DS_by_recurrence);

for(i=0,50,print(D_matrix^i*DS_samples));

vec_is_all_positive(v) = sum(i=1,#v, v[i]<0) == 0;
vec_sum_goes_negative(v) =
{
  \\ v=Vecrev(v);
  my(total=0);
  for(i=1,#v,
         total += v[i];
         if(total < 0, return(i)));
  return(0);
}

D_rec = [-1, 5, -9, 9, -8, 0, 16, -24, 43, -40, 16];
vecsum(D_rec)

D_vector = vector(50,i, if(i<=#D_rec, D_rec[i], 0));
add_at(pos) = my(f=D_vector[pos]); for(i=1,#D_rec, D_vector[pos+i-1] += f*D_rec[i]);
{
for(pos=2,#D_vector-#D_rec,
   my(neg=vec_sum_goes_negative(D_vector));
   print("neg="neg);
   add_at(pos));
}
{
  my(v=D_vector[-13..-1]);
  v = apply((x)->Str(x),v);
  len = vecmax(apply(length,v));
  for(i=1,#v,
     printf("%*s\n", len,v[i]));
}


gf_terms(g,n) = my(x=variable(g)); \
  concat(vector(min(n,valuation(g,x)), i, 0), \
         Vec(g + O(x^n)));
gf_term(g,n) = my(v=gf_terms(g,n+1)); v[#v];

{
  my(g = gDS(x));
  for(i=0,50,
     print(numerator(g));
     my(c = gf_term(g,0));
     g = (g - c)/x);
}
