\\ Copyright 2014, 2015, 2017 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file.  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.


default(recover,0);
read("c-curve-defines.gp");
gf_term(g,n) = my(v=gf_terms(g,n+1)); v[#v];

check_count = 0;
check_exact(got,want,name="") =
{
  check_count++;
  if(got!= want,
     print("wrong "name);
     print("got  "got);
     print("want "want);
     quit(1),
     print("ok "check_count" "name));
}

differ(x,y) =
{
  \\ if(type(x)!=type(y),
  \\    return(Strprintf("different types %s %s",type(x),type(y))));
  if(type(x)=="t_COMPLEX" || type(y)=="t_COMPLEX",
     my(d=differ(real(x),real(y)));
     if(d != 0, return(Strprintf("real %s",d)));
     d=differ(imag(x),imag(y));
     if(d != 0, return(Strprintf("imag %s",d)));
     return(0));

  if(type(x)=="t_VEC" || type(x)=="t_COL",
     if(length(x)!=length(y),
        return(Strprintf("different lengths %s %s",length(x),length(y))));
     for(i=1,#x,
        my(d=differ(x[i],y[i]));
        if(d != 0, return(Strprintf("at index %d %s",i,d))));
     return(0));

  if(! nearly_equal(x,y),
     return(Strprintf("different values\ngot  %s\nwant %s",x,y)));
  return(0);
}

check_equal(got,want,name="") =
{
  check_count++;
  my(d = differ(got,want));
  if(d != 0,
     print("wrong ",name,"\n",d); quit(1),
     print("ok ",check_count," ",name));
}

\\-----------------------------------------------------------------------------

\\   4      2      4        3      1      3
\\       ------->              <---------
\\   3      1      3        4      2      4

QR1(k) = TR(k,1) + 2*TR(k,8) + TR(k,9);
QR2(k) = TR(k,2) + 2*TR(k,3);
QR3(k) = TR(k,4) + TR(k,5);
QR4(k) = TR(k,6) + TR(k,7);
check_exact(1+2+1 + 1+2 + 2*2 + 2*2, 15)

QR_func_vector = [QR1,QR2,QR3,QR4];
check_exact(#QR_func_vector,4)
QR(k,j) = QR_func_vector[j](k);

check_exact(vector(#QR1_generated_samples,k,my(k=k-1); QR(k,1)), QR1_generated_samples, "QR1_generated_samples")
check_exact(vector(#QR2_generated_samples,k,my(k=k-1); QR(k,2)), QR2_generated_samples, "QR2_generated_samples")
check_exact(vector(#QR3_generated_samples,k,my(k=k-1); QR(k,3)), QR3_generated_samples, "QR3_generated_samples")
check_exact(vector(#QR4_generated_samples,k,my(k=k-1); QR(k,4)), QR4_generated_samples, "QR4_generated_samples")

QR_all_regions(k) = QR1(k) + QR2(k) + 2*QR3(k) + 2*QR4(k);
check_exact(vector(300,k, QR_all_regions(k)), vector(300,k, 2^k), \
            "QR_all_regions")
{
  print("QR");
  print1("      ");
  for(j=1,#QR_func_vector, printf("%7d", j) );
  print();

  for(k=0,20,
     printf("k=%2d  ", k);
     for(j=1,#QR_func_vector,
        printf("%7d", QR(k,j));
        );
     print());
}
print("QR1");
for(a=1,4, for(k=0,50,print1(QR(k,a),","));print());

gQR1(x) =
{
     + 14/35  * 1/(1 - 2*x)
     + 1/10 * (1 + 2*x)/(1 + x^2)
     + 1/2 * (1 + 2*x)/(1 - x^2 + 2*x^4);
}
check_exact(vector(1,a, gf_terms(gQR1(x),100)), \
            vector(1,a, vector(100,k,my(k=k-1); QR1(k))))
check_exact(1 - 4*1*2, -7)

gQR2(x) =
{
     + 1/35 * 1/(1 - 2*x)
     - 1/10 * (1 + 2*x)/(1 + x^2)
     + 1/14*(1 + 2*x - 4*x^2 - 8*x^3)/(1 - x^2 + 2*x^4);
}
check_exact(vector(1,a, gf_terms(gQR2(x),100)), \
            vector(1,a, vector(100,k,my(k=k-1); QR2(k))))

gQR3(x) =
{
     + 2/35 * 1/(1 - 2*x)
     + 1/20 * (1 + 2*x)/(1 + x^2)
     + 1/28*(-3 - 6*x - 2*x^2 - 4*x^3)/(1 - x^2 + 2*x^4);
}
check_exact(vector(1,a, gf_terms(gQR3(x),100)), \
            vector(1,a, vector(100,k,my(k=k-1); QR3(k))))

gQR4(x) =
{
     + 8/35 * 1/(1 - 2*x)
     - 1/20 * (1 + 2*x)/(1 + x^2)
     + 1/28*(-5 - 10*x + 6*x^2 + 12*x^3)/(1 - x^2 + 2*x^4);
}
check_exact(vector(1,a, gf_terms(gQR4(x),100)), \
            vector(1,a, vector(100,k,my(k=k-1); QR4(k))))


\\-----------------------------------------------------------------------------
\\ QR mutual recurrences

QR1_by_mutual(k) = if(k<2,QR1(k), \
  2*QR(k-2,1) + 4*QR(k-2,2) + 4*QR(k-2,3) + 2*QR(k-2,4));
check_exact(vector(100,k,my(k=k-1); QR1_by_mutual(k)), \
            vector(100,k,my(k=k-1); QR1(k)))

QR2_by_mutual(k) = if(k<2,QR2(k), 2*QR(k-2,3));
check_exact(vector(100,k,my(k=k-1); QR2_by_mutual(k)), \
            vector(100,k,my(k=k-1); QR2(k)))

QR3_by_mutual(k) = if(k<2,QR3(k), QR(k-2,4));
check_exact(vector(100,k,my(k=k-1); QR3_by_mutual(k)), \
            vector(100,k,my(k=k-1); QR3(k)))

QR4_by_mutual(k) = if(k<2,QR4(k), QR(k-2,1)   + QR(k-2,3) + 2*QR(k-2,4));
check_exact(vector(100,k,my(k=k-1); QR4_by_mutual(k)), \
            vector(100,k,my(k=k-1); QR4(k)))


\\-----------------------------------------------------------------------------
\\ QR by own recurrences

QR_by_recurrence(k,a) =
{
  if (k < 7, return(QR(k,a)));
    2* QR(k-1, a)
  -    QR(k-4, a)
  + 2* QR(k-5, a)
  - 2* QR(k-6, a)
  + 4* QR(k-7, a)
  ;
}

{
  for(a=1,4,
     check_exact(vector(12,k,k--; QR_by_recurrence(k,a)),
                 vector(12,k,k--; QR(k,a)),
                 Strprintf("a=%d",a)));
}

\\-----------------------------------------------------------------------------
\\ QR generating functions

QR_matrix = [2, 4, 4, 2; 0, 0, 2, 0; 0, 0, 0, 1; 1, 0, 1, 2];
QR_initial_even = [1;0;0;0];
QR_initial_odd  = [2;0;0;0];
QR_by_matrix(k,j) = (QR_matrix^floor(k/2) \
                     * if(k%2==0, QR_initial_even, QR_initial_odd))[j,1];
check_exact(matrix(200,#QR_func_vector,k,j, QR_by_matrix(k,j)), \
            matrix(200,#QR_func_vector,k,j, QR(k,j)));

QR_gf_column_even = ((matid(#QR_func_vector) - QR_matrix*x)^-1 * QR_initial_even)[,1];
{
  print("even");
  for(j=1,#QR_func_vector,
     pol_ascending_print(pol_partial_fractions(QR_gf_column_even[j])));
}
QR_gf_column_odd = ((matid(#QR_func_vector) - QR_matrix*x)^-1 * QR_initial_odd)[,1];
{
  print("odd");
  for(j=1,#QR_func_vector,
     pol_ascending_print(pol_partial_fractions(QR_gf_column_odd[j])));
}

\\ G - x^2*G*M = Initial
\\ G = (1 - x^2*M)^-1 * Initial
\\
QR_gf_column = ((matid(#QR_func_vector) - QR_matrix*x^2)^-1 \
                * (QR_initial_even + x*QR_initial_odd))[,1];
{
  print("all");
  for(s=1,#QR_gf_column,
     print("gQR" s "(x) = ", QR_gf_column[s]);
     print1("gQR" s "(x) = ");
     pol_ascending_print(pol_partial_fractions(QR_gf_column[s])));
}
check_exact(vector(1,a, gf_terms(QR_gf_column[a],100)), \
            vector(1,a, vector(100,k,my(k=k-1); QR(k,a))))
QR_by_gf_column(k,s) = gf_term(QR_gf_column[s], k);

{
  print("QR_by_gf_column");
  print1("      ");
  for(j=1,#QR_func_vector, printf("%7d", j) );
  print();

  for(k=1,16,
     printf("k=%2d  ", k);
     for(j=1,#QR_gf_column,
        printf("%7d", QR_by_gf_column(k,j));
        );
     print());
}
{
  for(k=2,20,
     print(QR1(k)/2^k * 1.0));
}
{
  for(k=2,20,
     print(gf_term(QR_gf_column[1],k) /2^k * 1.0));
}
