#!/usr/bin/perl -w

# Copyright 2015, 2017 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.


use 5.004;
use strict;
use FindBin;
use Math::PlanePath::CCurve 117; # v.117 for level_to_n_range()

use FindBin;
BEGIN { unshift @INC, $FindBin::Bin }
use MyCCurve 'xy_rotate_plus90', 'xy_rotate_minus90';


my $max_level = 14;
if (@ARGV) { $max_level = shift @ARGV; }
$| = 1;

print "% generated by tools/$FindBin::Script\n";
my $path = Math::PlanePath::CCurve->new;
my $var = $FindBin::Script;
$var =~ tr/-/_/;
$var =~ s/\.pl$//;

#         *-------*
#        /|\  9  /|\
#       / | \   / | \
#      /  |  \ /  |  \
#     *  7|8  * 10|11 *
#    / \  |  / \  |  / \
#   /   \ | /   \ | /   \
#  /  6  \|/  1  \|/ 12  \
# *-------*<<-----*-------*
#  \  5  /|\  2  /|\ 13  /
#   \   / | \   / | \   /
#    \ /  |  \ /  |  \ /
#     *  4|3  * 15|14 *
#      \  |  / \  |  /
#       \ | /   \ | /
#        \|/     \|/
#         *       *
#       bx2,by2

print "% GP-DEFINE  $var = [";
foreach my $level (0 .. $max_level) {
  my ($n_lo, $n_hi) = $path->level_to_n_range($level);

  my $check_xydxdy = sub {
    my ($x,$y, $dx,$dy) = @_;
    my $n = $path->xyxy_to_n($x,$y, $x+$dx,$y+$dy);
    return defined $n && $n < $n_hi;
  };
  my $check_xydxdy_both = sub {
    my ($x,$y, $dx,$dy) = @_;
    my @n_list = $path->xyxy_to_n_list_either($x,$y, $x+$dx,$y+$dy);
    @n_list = grep {$_ < $n_hi} @n_list;
    return @n_list == 2;
  };

  my $enc = 0;
  foreach my $n ($n_lo .. $n_hi-1) {
    my ($x,$y) = $path->n_to_xy($n);
    my ($dx,$dy) = $path->n_to_dxdy($n);

    my ($nx,$ny) = (-$dx, -$dy);
    my ($x2,$y2) = ($x+$dx, $y+$dy);
    my ($rx,$ry) = xy_rotate_minus90($dx,$dy);
    my ($lx,$ly) = xy_rotate_plus90($dx,$dy);

    # x,y, dx,dy  part 1
    $check_xydxdy_both->($x,$y, $rx,$ry) || next;       # part 10, 11
    $check_xydxdy_both->($x,$y, $nx,$ny) || next;       # part 12, 13
    $check_xydxdy_both->($x,$y, $lx,$ly) || next;       # part 14, 15

    $check_xydxdy_both->($x2,$y2, $dx,$dy) || next;     # part 5, 6
    $check_xydxdy_both->($x2,$y2, $rx,$ry) || next;     # part 7, 8
    $check_xydxdy     ->($x2,$y2, $nx,$ny) || next;     # part    2
    $check_xydxdy_both->($x2,$y2, $lx,$ly) || next;     # part 3, 4

    my ($tx2,$ty2) = ($x2+$rx, $y2+$ry);
    $check_xydxdy->($tx2,$ty2, $nx,$ny) || next;        # part 9

    # 2+2+2 + 2+2+1+2 + 1 == 14

    $enc++;
  }
  my $join = ($level == $max_level ? '' : ',');
  print $enc, $join;
}
print "];\n";

print "% end generated\n";
exit 0;
