#!/usr/bin/perl -w

# Copyright 2014, 2015, 2017 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.


use 5.004;
use strict;
use FindBin;
use List::Util 'min','max';
use Math::BigRat;
use Math::PlanePath::CCurve 117; # v.117 for level_to_n_range()

use FindBin;
BEGIN { unshift @INC, $FindBin::Bin }
use MyCCurve ':all';
use MyLineSegments;
use MyPrintwrap ':all';


print "  % generated by tools/$FindBin::Script\n";

my $level = 7;

my $path = Math::PlanePath::CCurve->new;
my ($n_lo, $n_hi) = $path->level_to_n_range($level);
my ($x_hi,$y_hi) = $path->n_to_xy($n_hi);
my ($mx,$my) = ($x_hi/2, $y_hi/2);

my $segments = MyLineSegments->new;
my $x_total = 0;
my $y_total = 0;
my $count = 0;
foreach my $n ($n_lo .. $n_hi-1) {
  my ($x,$y) = $path->n_to_xy($n);
  my ($dx,$dy) = $path->n_to_dxdy($n);
  $x_total += $x + $dx/2;
  $y_total += $y + $dy/2;
  $count++;

  $segments->add($x,$y, $x+$dx,$y+$dy);
}
my $gx = $x_total / $count;
my $gy = $y_total / $count;
my ($Gx,$Gy) = G_xy($level);

$Printwrap = 2;
printwrap("\\draw");
$Printwrap = 4;
foreach my $line ($segments->lines) {
  my ($x1,$y1, $x2,$y2) = @$line;
  printwrap(" ($x1,$y1)--($x2,$y2)");
}
printwrap(";\n");

my ($X_axis_x,$X_axis_y) = ($gx + $x_hi, $gy + $y_hi);
my ($Y_axis_x,$Y_axis_y) = ($gx + $y_hi, $gy - $x_hi);

print "  \\coordinate (centroid) at ($gx,$gy);\n";
print "  \\coordinate (start) at (0,0);\n";
print "  \\coordinate (end) at ($x_hi,$y_hi);\n";
print "  \\coordinate (X-axis) at ($X_axis_x,$X_axis_y);\n";
print "  \\coordinate (Y-axis) at ($Y_axis_x,$Y_axis_y);\n";
print "  % end generated\n";
exit 0;
