#!/usr/bin/perl -w

# Copyright 2017 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.

use 5.004;
use strict;
use FindBin;
use Lingua::EN::Numbers;
use Tie::IxHash;

use FindBin;
BEGIN { unshift @INC, $FindBin::Bin }
use MyCCurve 'div';
use MyPrintwrap ':all';


my $width = 14;
my $height = 8;

print "  % generated by tools/$FindBin::Script\n";

sub xy_to_turn {
  my ($x,$y) = @_;
  my $turn = 1;
  while (($x+$y) % 2 == 0) {
    ($x,$y) = div($x,$y); $x/=2; $y/=2;
    $turn--;
  }
  return $turn;
}

tie my %colours, 'Tie::IxHash';
foreach my $x (-$width .. $width) {
  foreach my $y (reverse -$height .. $height) {
    next if $x == 0 && $y == 0;
    my $turn = xy_to_turn($x,$y);
    push @{$colours{$turn}}, "$x,$y";
  }
}

foreach my $turn (sort {$b<=>$a} keys %colours) {
  my $style = sprintf "my turn%2d", $turn;
  $Printwrap = 2;
  printwrap("  \\fill[$style] foreach \\p in {");
  $Printwrap += 2;
  foreach my $xy (@{$colours{$turn}}) {
    printwrap("($xy),");
  }
  printwrap("}");
  printwrap(" { \\p +(-.5, -.5) rectangle +(.5, .5) };\n");
}

print "  % end generated\n";
exit 0;
